from dataclasses import dataclass, field

from beet import Context

from mecha import AstNode

from .provider import *


@dataclass
class AegisFeatureProviders:
    ctx: Context
    _providers: dict[type[AstNode], type[BaseFeatureProvider]] = field(
        init=False, default_factory=dict
    )

    def attach(self, node_type: type[AstNode], provider: type[BaseFeatureProvider]):
        self._providers[node_type] = provider

    def retrieve(
        self,
        node_type: type[AstNode] | AstNode,
    ) -> type[BaseFeatureProvider]:
        if not isinstance(node_type, type):
            node_type = type(node_type)

        return self._providers.get(node_type, BaseFeatureProvider)



from dataclasses import dataclass, field
from typing import Generic, TypeVar

import lsprotocol.types as lsp
from beet import Context
from mecha import AstNode

from ...semantics import TokenModifier, TokenType

__all__ = [
    "BaseFeatureProvider",
    "BaseParams",
    "CompletionParams",
    "HoverParams",
    "DefinitionParams",
    "ReferencesParams",
    "RenameParams",
    "SemanticsParams",
]

Node = TypeVar("Node", bound=AstNode)


@dataclass
class BaseParams(Generic[Node]):
    ctx: Context
    node: Node
    resource_location: str


@dataclass
class CompletionParams(BaseParams[Node]): ...


@dataclass
class HoverParams(BaseParams[Node]):
    text_range: lsp.Range


@dataclass
class DefinitionParams(BaseParams[Node]): ...


@dataclass
class ReferencesParams(BaseParams[Node]): ...


@dataclass
class RenameParams(BaseParams[Node]): ...


@dataclass
class SemanticsParams(BaseParams[Node]): ...


class BaseFeatureProvider(Generic[Node]):

    @classmethod
    def completion(
        cls, params: CompletionParams[Node]
    ) -> lsp.CompletionList | None:
        return None

    @classmethod
    def hover(cls, params: HoverParams[Node]) -> lsp.Hover | None:
        return None

    @classmethod
    def definition(
        cls, params: DefinitionParams[Node]
    ) -> list[lsp.Location | lsp.LocationLink] | lsp.Location | lsp.LocationLink | None:
        return None

    @classmethod
    def references(cls, params: ReferencesParams[Node]) -> list[lsp.Location] | None:
        return None

    @classmethod
    def rename(cls, params: RenameParams[Node]) -> lsp.WorkspaceEdit | None:
        return None

    @classmethod
    def semantics(
        cls, params: SemanticsParams[Node]
    ) -> list[tuple[AstNode, TokenType, list[TokenModifier]]] | None:
        return None



import inspect
import types
import typing
from copy import copy
from dataclasses import dataclass, field, fields, is_dataclass
from typing import Any, get_args, get_origin

UNKNOWN_TYPE = object()
TYPE_TO_INFO: dict[type, "TypeInfo"] = {}


@dataclass
class ParameterInfo:
    annotation: Any
    default: Any


@dataclass
class FunctionInfo:
    parameters: list[tuple[str, ParameterInfo]]
    return_annotation: Any
    doc: str | None

    @staticmethod
    def from_signature(
        signature: inspect.Signature, doc: str | None, hints: dict[str, Any]
    ) -> "FunctionInfo":
        return FunctionInfo(
            parameters=[
                (
                    name,
                    ParameterInfo(
                        annotation=parameter.annotation, default=parameter.default
                    ),
                )
                for name, parameter in signature.parameters.items()
            ],
            return_annotation=hints.get("return") or signature.return_annotation,
            doc=doc,
        )

    @staticmethod
    def extract(field_value: Any):
        try:
            hints = typing.get_type_hints(field_value)
            signature = inspect.signature(field_value)
            return FunctionInfo.from_signature(signature, field_value.__doc__, hints)
        except Exception as e:
            return FunctionInfo(
                parameters=[
                    (
                        "???",
                        ParameterInfo(inspect.Parameter.empty, inspect.Parameter.empty),
                    )
                ],
                return_annotation=UNKNOWN_TYPE,
                doc="Error while extracting signature: " + str(e),
            )

    def __hash__(self) -> int:
        return hash(self.__repr__())


@dataclass
class TypeInfo:
    doc: str | None
    fields: dict[str, Any] = field(default_factory=dict)
    functions: dict[str, FunctionInfo] = field(default_factory=dict)

    def get_member(self, field_name: str) -> Any | FunctionInfo:
        return self.fields.get(field_name) or self.functions.get(field_name)

    def add_member(
        self,
        field_annotations: dict[str, Any],
        field_name: str,
        field_value: Any,
        skip_fields=False,
    ):
        if (
            inspect.isfunction(field_value)
            or inspect.ismethod(field_value)
            or inspect.ismethoddescriptor(field_value)
            or inspect.isbuiltin(field_value)
        ):
            self.functions[field_name] = FunctionInfo.extract(field_value)

        elif skip_fields:
            return
        elif field_name in field_annotations:
            self.fields[field_name] = field_annotations[field_name]
        elif _ := get_origin(field_value):
            self.fields[field_name] = field_value
        else:
            self.fields[field_name] = type(field_value)

    def __hash__(self) -> int:
        return hash(self.__repr__())


def get_type_info(_type: type) -> TypeInfo:

    if _type in TYPE_TO_INFO:
        return TYPE_TO_INFO[type]

    info = TypeInfo(_type.__doc__)

    # logging.debug("\n\n")
    # logging.debug("-" * 50)
    # logging.debug("Indexing type " + _type.__name__)

    field_annotations = (
        inspect.get_annotations(_type)
        if inspect.isclass(_type) or callable(_type) or inspect.ismodule(_type)
        else {}
    )

    if is_dataclass(_type):
        handled_fields = set()
        for field in fields(_type):
            info.add_member({}, field.name, field.type)
            handled_fields.add(field.name)
        for field_name, field_value in inspect.getmembers(_type):
            if field_name in handled_fields:
                continue

            # field_value = getattr(_type, field_name)
            info.add_member(field_annotations, field_name, field_value)
    else:
        for field_name, field_value in inspect.getmembers(_type):
            # field_value = getattr(_type, field_name)
            # logging.debug(f"{field_name}, {field_value}")
            info.add_member(field_annotations, field_name, field_value)

    # logging.debug("-" * 50)
    # logging.debug("\n\n")
    # logging.debug(info)
    return info


def get_name_of_type(annotation):
    if annotation is inspect.Parameter.empty:
        return None

    if annotation is types.NoneType:
        return "None"

    if annotation is UNKNOWN_TYPE:
        return "???"

    origin = typing.get_origin(annotation)
    args = list(map(str, map(get_name_of_type, typing.get_args(annotation))))

    if origin is typing.Union or origin is types.UnionType:
        return " | ".join(args)
    elif origin:
        return f"{origin.__name__}[{', '.join(args)}]"

    try:
        return annotation.__repr__()
    except:
        return annotation.__name__ if hasattr(annotation, "__name__") else annotation


def format_function_hints(
    name: str,
    signature: FunctionInfo,
    keyword: str = "def",
    show_return_type: bool = True,
):
    hint = f"{keyword} {name}("

    return_type = signature.return_annotation

    parameters = []

    for name, parameter in signature.parameters:
        annotation = get_name_of_type(parameter.annotation)

        if annotation is None and parameter.default is not inspect.Parameter.empty:
            annotation = get_name_of_type(type(parameter.default))

        annotation_string = ": " + str(annotation) if annotation else ""
        default_string = (
            " = " + parameter.default.__repr__()  # type: ignore
            if parameter.default is not inspect.Parameter.empty
            else ""
        )

        parameters.append(f"\n\t{name}{annotation_string}{default_string}")

    hint += ",".join(parameters)

    hint += f"\n)"

    if show_return_type and signature.return_annotation is not types.NoneType:
        hint += f" -> {get_name_of_type(return_type)}"

    return hint


def get_doc_string(doc: Any):
    return "\n---\n" + doc if isinstance(doc, str) else ""


def get_variable_description(name: str, value: Any):
    if inspect.isclass(value):
        return f"```python\n(variable) {name}: {get_name_of_type(value)}\n```"

    doc_string = get_doc_string(value.__doc__)
    return f"```python\n(variable) {name}: {get_name_of_type(type(value))}\n```{doc_string}"


def get_class_description(name: str, value: type | TypeInfo):
    if not isinstance(value, TypeInfo):
        value = get_type_info(value)

    doc_string = get_doc_string(value.doc)

    if not (init := value.functions.get("__init__")):
        return f"```python\nclass {name}()\n```{doc_string}"

    init = copy(init)

    if len(init.parameters) > 0:
        init.parameters.pop(0)

    return f"```python\n{format_function_hints(name, init, keyword='class', show_return_type=False)}\n```{doc_string}"


def get_function_description(name: str, function: Any):
    function_info = None
    if isinstance(function, FunctionInfo):
        function_info = function
    else:
        function_info = FunctionInfo.extract(function)

    doc_string = get_doc_string(function_info.doc)

    return f"```py\n{format_function_hints(name, function_info)}\n```{doc_string}"


def get_annotation_description(name: str, type_annotation: Any):
    if get_origin(type_annotation) is type:
        args = get_args(type_annotation)
        description = get_class_description(name, args[0])
    elif isinstance(type_annotation, TypeInfo):
        description = get_class_description(name, type_annotation)
    elif (
        inspect.isfunction(type_annotation)
        or inspect.isbuiltin(type_annotation)
        or isinstance(type_annotation, FunctionInfo)
    ):
        description = get_function_description(name, type_annotation)
    else:
        description = get_variable_description(name, type_annotation)

    return description



from typing import Literal, Union

TokenModifier = Union[
    Literal["declaration"],
    Literal["definition"],
    Literal["readonly"],
    Literal["static"],
    Literal["deprecated"],
    Literal["abstract"],
    Literal["async"],
    Literal["modification"],
    Literal["documentation"],
    Literal["defaultLibrary"],
]

TokenType = Union[
    Literal["comment"],
    Literal["string"],
    Literal["keyword"],
    Literal["number"],
    Literal["regexp"],
    Literal["operator"],
    Literal["namespace"],
    Literal["type"],
    Literal["struct"],
    Literal["class"],
    Literal["interface"],
    Literal["enum"],
    Literal["typeParameter"],
    Literal["function"],
    Literal["method"],
    Literal["decorator"],
    Literal["macro"],
    Literal["variable"],
    Literal["parameter"],
    Literal["property"],
    Literal["label"],
]



from typing import Iterable

import lsprotocol.types as lsp
from mecha import AstNode
from tokenstream import SourceLocation


def node_location_to_range(node: AstNode | Iterable[SourceLocation]):
    if isinstance(node, AstNode):
        location = node.location
        end_location = node.end_location
    else:
        location, end_location = node

    return lsp.Range(
        start=location_to_position(location), end=location_to_position(end_location)
    )


def node_start_to_range(node: AstNode):
    start = location_to_position(node.location)
    end = lsp.Position(line=start.line, character=start.character + 1)

    return lsp.Range(start=start, end=end)


def location_to_position(location: SourceLocation) -> lsp.Position:
    return lsp.Position(
        line=max(location.lineno - 1, 0),
        character=max(location.colno - 1, 0),
    )


def offset_location(location: SourceLocation, offset):
    return SourceLocation(
        location.pos + offset, location.lineno, location.colno + offset
    )



import logging
import re
from dataclasses import dataclass, field
from pathlib import Path
from threading import Lock
from typing import ClassVar

from beet import Context, File, NamespaceFile
from beet.core.utils import extra_field, required_field
from tokenstream import SourceLocation

__all__ = ["FilePointer", "ResourceIndex", "AegisProjectIndex"]

FilePointer = tuple[SourceLocation, SourceLocation]


@dataclass
class ResourceIndice:
    definitions: dict[str, set[FilePointer]] = extra_field(default_factory=dict)
    references: dict[str, set[FilePointer]] = extra_field(default_factory=dict)

    def _dump(self) -> str:
        dump = ""

        dump += "definitions:\n"
        for path, pointers in self.definitions.items():
            for pointer in pointers:
                dump += f"\t- {path} {pointer[0].lineno}:{pointer[0].colno} -> {pointer[1].lineno}:{pointer[1].colno}\n"
        dump += "references:\n"
        for path, pointers in self.references.items():
            for pointer in pointers:
                dump += f"\t- {path} {pointer[0].lineno}:{pointer[0].colno} -> {pointer[1].lineno}:{pointer[1].colno}\n"

        return dump


def valid_resource_location(path: str):
    return bool(re.match(r"^[a-z0-9_\.]+:[a-z0-9_\.]+(\/?[a-z0-9_\.]+)*$", path))


@dataclass
class ResourceIndex:
    _files: dict[str, ResourceIndice] = extra_field(default_factory=dict)
    _lock: Lock = extra_field(default_factory=Lock)

    def remove_associated(self, path: str | File):
        self._lock.acquire()

        if isinstance(path, File):
            path = str(Path(path.ensure_source_path()).absolute())

        for file, indice in list(self._files.items()):
            if path in indice.definitions:
                del indice.definitions[path]
            if path in indice.references:
                del indice.references[path]

            if len(indice.definitions) == 0:
                del self._files[file]

        self._lock.release()

    def add_definition(
        self,
        resource_path: str,
        source_path: str,
        source_location: FilePointer = (
            SourceLocation(0, 0, 0),
            SourceLocation(0, 0, 0),
        ),
    ):
        if not valid_resource_location(resource_path):
            raise Exception(f"Invalid resource location {resource_path}")

        self._lock.acquire()

        indice = self._files.setdefault(resource_path, ResourceIndice())
        locations = indice.definitions.setdefault(source_path, set())
        locations.add(source_location)

        self._lock.release()

    def get_definitions(
        self, resource_path: str
    ) -> list[tuple[str, SourceLocation, SourceLocation]]:
        if not (file := self._files.get(resource_path)):
            return []

        definitions = []
        for path, locations in file.definitions.items():
            for location in locations:
                definitions.append((path, *location))

        return definitions

    def get_references(
        self, resource_path: str
    ) -> list[tuple[str, SourceLocation, SourceLocation]]:
        if not (file := self._files.get(resource_path)):
            return []

        references = []
        for path, locations in file.references.items():
            for location in locations:
                references.append((path, *location))

        return references

    def add_reference(
        self,
        resource_path: str,
        source_path: str,
        source_location: FilePointer = (
            SourceLocation(0, 0, 0),
            SourceLocation(0, 0, 0),
        ),
    ):
        if not valid_resource_location(resource_path):
            raise Exception(f"Invalid resource location {resource_path}")

        self._lock.acquire()

        indice = self._files.setdefault(resource_path, ResourceIndice())
        locations = indice.references.setdefault(source_path, set())
        locations.add(source_location)

        self._lock.release()

    def __iter__(self):
        items = self._files.keys()

        for item in items:
            yield item

    def _dump(self) -> str:
        dump = ""

        for file, indice in self._files.items():
            dump += f"\n- '{file}':\n"
            dump += "\t" + "\n\t".join(indice._dump().splitlines())

        return dump


@dataclass
class AegisProjectIndex:
    _ctx: Context
    _resources: dict[type[NamespaceFile], ResourceIndex] = field(default_factory=dict)

    resource_name_to_type: dict[str, type[NamespaceFile]] = field(default_factory=dict)

    def __post_init__(self):
        self.resource_name_to_type = {
            t.snake_name: t for t in self._ctx.get_file_types()
        }

    def __getitem__(self, key: type[NamespaceFile]):
        return self._resources.setdefault(key, ResourceIndex())

    def remove_associated(self, path: str):
        for resource in self._resources.values():
            resource.remove_associated(path)

    def dump(self) -> str:
        dump = ""
        for resource, index in self._resources.items():
            dump += f"\nResource {resource.__name__}:"
            dump += "\t" + "\n\t".join(index._dump().splitlines())

        return dump



from dataclasses import dataclass, field
from typing import Any, TypeVar

from beet import NamespaceFile
from mecha import AstNode

from ..reflection import UNKNOWN_TYPE

METADATA_KEY = "aegis_metadata"

__all__ = [
    "BaseMetadata",
    "VariableMetadata",
    "ResourceLocationMetadata",
    "attach_metadata",
    "retrieve_metadata",
]


@dataclass
class BaseMetadata:
    """
    BaseMetadata provides information to aegis_server about the AstNode.
    """


@dataclass
class VariableMetadata(BaseMetadata):
    """
    VariableMetadata provides information to aegis_server about a node representing a Bolt variable

    Attributes
    ----------
    type_annotation : Any
        The python type that the node represents

    documentation : str
        The documentation string for the node
    """

    type_annotation: Any = field(default=UNKNOWN_TYPE)

    documentation: str | None = field(default=None)


@dataclass
class ResourceLocationMetadata(BaseMetadata):
    """
    ResourceLocationMetadata provides information to aegis_server about a node representing a resource location node

    Attributes
    ----------
    respresents : str | type[NamespaceFile] | None
        The registry or type of File the resource location represents

    unresolved_path : str | None
        The unresolved path of the string, ex. ~/foo
    """

    represents: str | type[NamespaceFile] | None = field(default=None)

    unresolved_path: str | None = field(default=None)


def attach_metadata(node: AstNode, metadata: BaseMetadata):
    """
    Attaches the provided metadata instance to the node

    Parameters
    ----------
    node : AstNode
        The node to attach the metadata too
    metadata : BaseMetadata
        The metadata to be attached
    """
    node.__dict__[METADATA_KEY] = metadata


T = TypeVar("T")


def retrieve_metadata(
    node: AstNode, type: tuple[type[T]] | type[T] = BaseMetadata
) -> T | None:
    """
    Retrieves the metadata attached to a node

    Parameters
    ----------
    node : AstNode
        The node to retrieve from
    type : tuple[type] | type
        The type to check the metadata for

    Returns
    -------
    BaseMetadata
        The metadata attached to the node
    None
        If not metadata is present on the node
    """
    metadata = node.__dict__.get(METADATA_KEY)

    if isinstance(metadata, type):
        return metadata

    return None



from dataclasses import dataclass, field

from beet import Context


@dataclass
class AegisGameRegistries:
    ctx: Context

    registries: dict[str, list[str]] = field(init=False, default_factory=dict)

    def __getitem__(self, registry: str) -> list[str]:
        return self.registries.get(registry) or []

    def __contains__(self, registry: str) -> bool:
        return registry in self.registries
